#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute the (non)linear flow around extruded 2D cylinder, with various B.C.
# Adrien Crovato
#
# Test solver convergence and boundary conditions
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import math
import dart.default as floD
import fwk
from fwk.testing import *
from fwk.coloring import ccolors 

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 3*math.pi/180
    U_inf = [math.cos(alpha), 0., math.sin(alpha)] # norm should be 1
    M_inf = 0.1
    dim = 3

    # define dimension and mesh size
    dmt = 1.0 # cylinder diameter
    lgt = 8.0 # channel length
    hgt = 6.0 # channel height
    wdt = 3.0 # channel width
    l_ref = dmt # reference length
    S_ref = dmt*wdt # reference area
    fms = 0.5 # farfield mesh size
    nms = 0.1 # nearfield mesh size  

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms["msh"].start()
    pars = {'dmt' : dmt, 'lgt' : lgt, 'wdt' : wdt, 'hgt' : hgt, 'msN' : nms, 'msF' : fms}
    msh, gmshWriter = floD.mesh(dim, 'models/cylinder_2D5.geo', pars, ['field', 'cylinder', 'symmetry', 'downstream'])
    tms["msh"].stop()

    # set the problem and add fluid, initial/boundary conditions
    tms['pre'].start()
    pbl, dirichlet, wake, bnd, _ = floD.problem(msh, dim, alpha, 0., M_inf, S_ref, l_ref, 0., 0., 0., 'cylinder', tp = 'te', dbc=True)
    tms['pre'].stop()

    # solve problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    solver = floD.newton(pbl)
    cnvrgd = solver.run()
    solver.save(gmshWriter)
    tms['solver'].stop()

    # compute mean error in Body, Neumann boundary condition (works only for Tri3 elements)
    gradn = []
    for e in bnd.groups[0].tag.elems:
        gradx = 0
        grady = 0
        gradz = 0
        ux = e.nodes[1].pos[0] - e.nodes[0].pos[0]
        uy = e.nodes[1].pos[1] - e.nodes[0].pos[1]
        uz = e.nodes[1].pos[2] - e.nodes[0].pos[2]
        vx = e.nodes[2].pos[0] - e.nodes[0].pos[0]
        vy = e.nodes[2].pos[1] - e.nodes[0].pos[1]
        vz = e.nodes[2].pos[2] - e.nodes[0].pos[2]
        nx = uy*vz - uz*vy
        ny = uz*vx - ux*vz
        nz = ux*vy - uy*vx
        for n in e.nodes:
            gradx += solver.U[n.row][0] / e.nodes.size()
            grady += solver.U[n.row][1] / e.nodes.size()
            gradz += solver.U[n.row][2] / e.nodes.size()
        gradn_ = gradx * nx + grady * ny + gradz * nz
        gradn.append(gradn_)
    errNeu = sum(gradn) / len(gradn)
    # compute largest error in Dirichlet boundary condition
    errDir = 0.
    for n in dirichlet.nodes:
        errCur = solver.phi[n.row] - (n.pos[0]*U_inf[0] + n.pos[1]*U_inf[1] + n.pos[2]*U_inf[2])
        if (abs(errCur) > errDir):
            errDir = abs(errCur)
    # compute largest error in Wake boundary condition
    errWak = 0.
    for n in wake.nodMap:
        errCur = solver.Cp[n.row] - solver.Cp[wake.nodMap[n].row]
        if (abs(errCur) > errWak):
            errWak = abs(errCur)

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # visualize solution
    floD.initViewer(pbl)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    if (not cnvrgd):
        raise Exception(ccolors.ANSI_RED + 'Flow solver failed to converge!' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('iteration count', solver.nIt, 3, 1, forceabs=True))
    tests.add(CTest('Neumann B.C. mean error', errNeu, 0., 1e-2))
    tests.add(CTest('Dirichlet B.C. max error', errDir, 0., 1e-12))
    tests.add(CTest('Wake/Kutta B.C. max error', errWak, 0., 1e-1))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
