################################################################################
#                                                                              #
#                           DARTFlo viscous module file                        #
#                    Integration: Newton method for convergence								 #
# 															to steady state.                   						 #
#                                                                              #
# Author: Paul Dechamps                                                        #
# Université de Liège                                                          #
# Date: 2021.09.10                                                             #
#                                                                              #
################################################################################


# ------------------------------------------------------------------------------
#  Imports
# ------------------------------------------------------------------------------
from matplotlib.colors import Normalize
from dart.viscous.Physics.DClosures import Closures

import math
import numpy as np
from matplotlib import pyplot as plt
from fwk.coloring import ccolors


class Integration:
  """Pseudo-time integration of the integral boundary layer equations.
  The non-linear equations are solved for one point using Newton's method
  
  Attributes
  ----------
  - CFL0 (float): Starting CFL.
  - maxIt (int): Maximum number of iterations.
  - tol (float): Convergence criterion.
  - itFrozenJac0 (int): Number of iterations during which the Jacobian matrix is frozen.
  - Minf (float): Free-stream Mach number.
  - gamma (float): Fluid heat capacity ratio.
  """

  def __init__(self, _sysMatrix, _Minf, _Cfl0) -> None:
    self.sysMatrix = _sysMatrix

    self.__CFL0=_Cfl0
    self.__maxIt = 1000
    self.__tol = 1e-8
    self.itFrozenJac0 = 5
    
    self.__Minf = max(_Minf, 0.1)
    self.gamma = 1.4
    pass

  def GetCFL0(self): return self.__CFL0
  def SetCFL0(self, _cfl0): self.__CFL0 = _cfl0

  def Gettol(self): return self.__tol
  def Settol(self, _tol): self.__tol = _tol

  def GetmaxIt(self): return self.__maxIt
  def SetmaxIt(self, _maxIt): self.__maxIt = _maxIt

  def TimeMarching (self, cfgFlow, iMarker, screenOutput = 0) -> int:
    """Pseudo-time marching solver.
    
    Args
    ----
    - DFlow (class type): Data structure containing the flow configuration.
    - iMarker (int): Id of the point to be treated.
    
    Returns
    -------
    - Exit code info (int):	
      -  0 -> sucessfull exit.
      - >0 -> convergence to tolerance not achieved, number of iterations.
      - <0 -> illegal input or breakdown.
    
    Opts
    ----
    - screenOutput (bool, optional): Output pseudo-time marching convergence.
    """

    # Save initial condition.

    sln0 = cfgFlow.u.copy()

    # Retreive parameters.

    nVar = cfgFlow.nVar                                             # Number of variables of the differential problem.
    nExtPar = cfgFlow.nExtPar                                       # Number of boundary layer parameters used for vector indexing.
    bNodes = cfgFlow.bNodes                                         # Boundary nodes of the computational domain.
    iInvVel = cfgFlow.extPar[iMarker*nExtPar + 2]                   # Inviscid velocity on the considered cell.
    iMarkerPrev = 0 if iMarker == bNodes[1] else iMarker - 1      	# Point upstream of the considered point.
    dx = cfgFlow.xx[iMarker] - cfgFlow.xx[iMarkerPrev]              # Cell size.

    # Initial residual.

    sysRes = self.sysMatrix.buildResiduals(cfgFlow, iMarker)
    normSysRes0 = np.linalg.norm(sysRes)

    # Initialize pseudo-time integration parameters.

    CFL = self.__CFL0                                             # Initialized CFL number.
    itFrozenJac = self.itFrozenJac0                               # Number of iterations for which Jacobian is frozen.
    timeStep = self.__SetTimeStep(CFL, dx, iInvVel)               # Initial time step.
    IMatrix = np.identity(nVar) / timeStep                        # Identity matrix used to damp the system.
    convPlot = []
    # Main loop.

    nErrors = 0                                                   # Count for the number of errors identified.
    innerIters = 0                                                # Number of inner iterations to solver the non-linear system.
    while innerIters <= self.__maxIt:
      try:
        if iMarker == 2:
          quit()
        # Jacobian computation.

        if innerIters % itFrozenJac == 0:
          Jacobian=self.sysMatrix.buildJacobian(cfgFlow, iMarker,
                                                sysRes, order = 1, eps = 1e-8)

        # Linear solver.

        slnIncr = np.linalg.solve((Jacobian + IMatrix), - sysRes)

        # Increment solution.

        cfgFlow.u[iMarker*nVar : (iMarker+1)*nVar] += slnIncr

        # Residual computation.

        sysRes = self.sysMatrix.buildResiduals(cfgFlow, iMarker)
        normSysRes = np.linalg.norm(sysRes + slnIncr/timeStep)
        print('CFL', CFL, 'timeStep', timeStep)
        convPlot.append(normSysRes/normSysRes0)

        # Check convergence.
        
        if normSysRes <= self.__tol * normSysRes0:
          # Convergence criterion satisfied.
          return 0

      except KeyboardInterrupt:
        print(ccolors.ANSI_RED + 'Program terminated by user.', ccolors.ANSI_RESET)
        quit()
        
      except Exception:
        print('it ', innerIters, ' Entered Exception')

        nErrors += 1

        if nErrors >= 5:
          self.__ResetSolution(cfgFlow, iMarker, sln0)
          return -1

        cfgFlow.u[iMarker*nVar : (iMarker+1)*nVar] = cfgFlow.u[(iMarker-1)*nVar : (iMarker)*nVar]		# Reset solution.

        # Lower CFL and recompute time step

        if math.isnan(CFL):
          CFL = self.__CFL0
        CFL = min(max(CFL / 2,0.3),1)
        timeStep = self.__SetTimeStep(CFL, dx, iInvVel)
        IMatrix = np.identity(nVar) / timeStep

        sysRes = self.sysMatrix.buildResiduals(cfgFlow, iMarker)

        itFrozenJac = 1
        continue

      # CFL adaptation.

      CFL = self.__CFL0* (normSysRes0/normSysRes)**0.7

      CFL = max(CFL, 1)
      timeStep = self.__SetTimeStep(CFL, dx, iInvVel)
      IMatrix = np.identity(nVar) / timeStep


      innerIters += 1

    else: # Maximum number of iterations reached.

      if normSysRes >= 1e-3 * normSysRes0:
        self.__ResetSolution(cfgFlow, iMarker, sln0)
      return innerIters


  def __GetSoundSpeed(self, invVelSquared) -> float:
    """Compute the speed of sound in a cell.

    Args
    ----
    - invVelSquared (float): Square of the inviscid velocity.
    """
    
    return np.sqrt(1 / (self.__Minf * self.__Minf) + 0.5 * (self.gamma - 1) * (1 - invVelSquared))

  def __SetTimeStep(self, CFL, dx, invVel) -> float:
    """Compute the pseudo-time step in a cell for a given
    CFL number.

    Args
    ----
    - CFL (float): Current CFL number.
    - dx (float): Cell size.
    - invVel (float): Inviscid velocity in the cell.

    Returns
    -------
    - timeStep (float): Pseudo-time step.
    """

    # Speed of sound.
    gradU2 = (invVel)**2
    # Velocity of the fastest travelling wave
    soundSpeed = self.__GetSoundSpeed(gradU2)
    advVel = soundSpeed + invVel

    # Time step computation 
    return CFL*dx/advVel

  def __ResetSolution(self, DFlow, iMarker, sln0) -> None:

    nVar 	 	= DFlow.nVar
    nBlPar 	= DFlow.nBlPar
    nExtPar = DFlow.nExtPar
    
    DFlow.u[iMarker*nVar:(iMarker+1)*nVar] = sln0[(iMarker)*nVar: (iMarker+1)*nVar].copy()
    name = 'airfoil' if iMarker <= DFlow.bNodes[2] else 'wake'

    if DFlow.flowRegime[iMarker]==0:

      # Laminar closure relations
      DFlow.blPar[iMarker*nBlPar+1:iMarker*nBlPar+7] = Closures.laminarClosure(DFlow.u[iMarker*nVar + 0], DFlow.u[iMarker*nVar + 1],
                                              DFlow.blPar[iMarker*nBlPar+0], DFlow.extPar[iMarker*nExtPar+0],
                                              name)

    elif DFlow.flowRegime[iMarker]==1:

      # Turbulent closures relations 
      DFlow.blPar[iMarker*nBlPar+1:iMarker*nBlPar+9] = Closures.turbulentClosure(DFlow.u[iMarker*nVar + 0], DFlow.u[iMarker*nVar + 1],
                                              DFlow.u[iMarker*nVar + 4], DFlow.blPar[iMarker*nBlPar+0],
                                              DFlow.extPar[iMarker*nExtPar+0], name)
